"""
ProbLog command-line interface.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

import stat
import sys
import os
import traceback

from ..program import PrologFile
from ..evaluator import SemiringLogProbability, SemiringProbability, SemiringSymbolic
from .. import get_evaluatable, get_evaluatables

from ..util import Timer, start_timer, stop_timer, init_logger, format_dictionary
from ..errors import process_error


def print_result(d, output, precision=8):
    """Pretty print result.

    :param d: result from run_problog
    :param output: output file
    :param precision:
    :return:
    """
    success, d = d
    if success:
        print(format_dictionary(d, precision), file=output)
        return 0
    else:
        print (process_error(d), file=output)
        return 1


def print_result_json(d, output, precision=8):
    """Pretty print result.

    :param d: result from run_problog
    :param output: output file
    :param precision:
    :return:
    """
    import json
    result = {}
    success, d = d
    if success:
        result['SUCCESS'] = True
        result['probs'] = [[str(n), round(p, precision), n.loc[1], n.loc[2]] for n, p in d.items()]
    else:
        result['SUCCESS'] = False
        result['err'] = vars(d)
        result['err']['message'] = str(process_error(d))
    print (json.dumps(result), file=output)
    return 0


def execute(filename, knowledge=None, semiring=None, debug=False, **kwdargs):
    """Run ProbLog.

    :param filename: input file
    :param knowledge: knowledge compilation class or identifier
    :param semiring: semiring to use
    :param parse_class: prolog parser to use
    :param debug: enable advanced error output
    :param engine_debug: enable engine debugging output
    :param kwdargs: additional arguments
    :return: tuple where first value indicates success, and second value contains result details
    """

    if knowledge is None or type(knowledge) == str:
        knowledge = get_evaluatable(knowledge)
    try:
        with Timer('Total time'):
            model = PrologFile(filename)
            formula = knowledge.create_from(model, **kwdargs)
            result = formula.evaluate(semiring=semiring, **kwdargs)

            # Update loceation information on result terms
            for n, p in result.items():
                if not n.location or not n.location[0]:
                    # Only get location for primary file (other file information is not available).
                    n.loc = model.lineno(n.location)
        return True, result
    except Exception as err:
        trace = traceback.format_exc()
        err.trace = trace
        return False, err


def argparser():
    """Create the default argument parser for ProbLog.
    :return: argument parser
    :rtype: argparse.ArgumentParser
    """
    import argparse

    class InputFile(str):
        """Stub class for file input arguments."""
        pass

    class OutputFile(str):
        """Stub class for file output arguments."""
        pass

    description = """ProbLog 2.1 command line interface

    The arguments listed below are for the default mode.
    ProbLog also supports the following alternative modes:

      - (default): inference
      - install: run the installer
      - explain: compute the probability of a query and explain how to get there
      - ground: generate ground program (see ground --help)
      - sample: generate samples from the model (see sample --help)
      - unittest: run the testsuite

    Select a mode by adding one of these keywords as first argument (e.g. problog-cli.py install).
    """

    parser = argparse.ArgumentParser(description=description,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('filenames', metavar='MODEL', nargs='*', type=InputFile)
    parser.add_argument('--verbose', '-v', action='count', help='Verbose output')
    parser.add_argument('--knowledge', '-k', dest='koption',
                        choices=get_evaluatables(),
                        default=None, help="Knowledge compilation tool.")

    # Evaluation semiring
    ls_group = parser.add_mutually_exclusive_group()
    ls_group.add_argument('--logspace', action='store_true',
                          help="Use log space evaluation (default).", default=True)
    ls_group.add_argument('--nologspace', dest='logspace', action='store_false',
                          help="Use normal space evaluation.")
    ls_group.add_argument('--symbolic', dest='symbolic', action='store_true',
                          help="Use symbolic computations.")

    parser.add_argument('--output', '-o', help="Output file (default stdout)", type=OutputFile)
    parser.add_argument('--recursion-limit',
                        help="Set Python recursion limit. (default: %d)" % sys.getrecursionlimit(),
                        default=sys.getrecursionlimit(), type=int)
    parser.add_argument('--timeout', '-t', type=int, default=0,
                        help="Set timeout (in seconds, default=off).")
    parser.add_argument('--compile-timeout', type=int, default=0,
                        help="Set timeout for compilation (in seconds, default=off).")
    parser.add_argument('--debug', '-d', action='store_true',
                        help="Enable debug mode (print full errors).")
    parser.add_argument('--web', action='store_true', help=argparse.SUPPRESS)

    # Additional arguments (passed through)
    parser.add_argument('--engine-debug', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('--propagate-evidence', action='store_true',
                        dest='propagate_evidence',
                        default=argparse.SUPPRESS,
                        help="Enable evidence propagation")
    parser.add_argument('--propagate-weights', action='store_true', default=None,
                        help="Enable weight propagation")
    parser.add_argument('--convergence', '-c', type=float, default=argparse.SUPPRESS,
                        help='stop anytime when bounds are within this range')

    # SDD garbage collection
    sdd_auto_gc_group = parser.add_mutually_exclusive_group()
    sdd_auto_gc_group.add_argument('--sdd-auto-gc', action='store_true', dest='sdd_auto_gc',
                                   default=argparse.SUPPRESS, help=argparse.SUPPRESS)
    sdd_auto_gc_group.add_argument('--sdd-no-auto-gc', action='store_false', dest='sdd_auto_gc',
                                   default=argparse.SUPPRESS, help=argparse.SUPPRESS)

    sdd_fixvars_group = parser.add_mutually_exclusive_group()
    sdd_fixvars_group.add_argument('--sdd-preset-variables', action='store_true',
                                   dest='sdd_preset_variables',
                                   default=argparse.SUPPRESS, help=argparse.SUPPRESS)
    sdd_fixvars_group.add_argument('--sdd-no-preset-variables', action='store_false',
                                   dest='sdd_preset_variables',
                                   default=argparse.SUPPRESS, help=argparse.SUPPRESS)

    return parser


def main(argv, result_handler=None):
    parser = argparser()
    args = parser.parse_args(argv)

    if result_handler is None:
        if args.web:
            result_handler = print_result_json
        else:
            result_handler = print_result

    init_logger(args.verbose)

    if args.output is None:
        output = sys.stdout
    else:
        output = open(args.output, 'w')

    if args.timeout:
        start_timer(args.timeout)

    if len(args.filenames) == 0:
        mode = os.fstat(0).st_mode
        if stat.S_ISFIFO(mode) or stat.S_ISREG(mode):
            # stdin is piped or redirected
            args.filenames = ['-']
        else:
            # stdin is terminal
            # No interactive input, exit
            print('ERROR: Expected a file or stream as input.\n', file=sys.stderr)
            parser.print_help()
            sys.exit(1)

    if args.logspace:
        semiring = SemiringLogProbability()
    else:
        semiring = SemiringProbability()

    if args.symbolic:
        args.koption = 'nnf'
        semiring = SemiringSymbolic()

    if args.propagate_weights:
        args.propagate_weights = semiring

    for filename in args.filenames:
        if len(args.filenames) > 1:
            print ('Results for %s:' % filename)
        result = execute(filename, args.koption, semiring, **vars(args))
        retcode = result_handler(result, output)
        if len(args.filenames) == 1:
            sys.exit(retcode)

    if args.output is not None:
        output.close()

    if args.timeout:
        stop_timer()


if __name__ == '__main__':
    main(sys.argv[1:])



"""

problog.logic - Basic logic
---------------------------

This module contains basic logic constructs.

    A Term can be:
        * a function (see :class:`Term`)
        * a variable (see :class:`Var`)
        * a constant (see :class:`Constant`)

    Four functions are handled separately:
        * conjunction (see :class:`And`)
        * disjunction (see :class:`Or`)
        * negation (see :class:`Not`)
        * clause (see :class:`Clause`)

    **Syntactic sugar**

    Clauses can be constructed by virtue of overloading of Python operators:

      =========== =========== ============
       Prolog      Python      English
      =========== =========== ============
       ``:-``          ``<<``      clause
       ``,``           ``&``       and
       ``;``           ``|``       or
       ``\+``          ``~``       not
      =========== =========== ============

    .. warning::

        Due to Python's operator priorities, the body of the clause has to be between parentheses.


    **Example**::

        from problog.logic import Var, Term

        # Functors (arguments will be added later)
        ancestor = Term('anc')
        parent = Term('par')

        # Literals
        leo3 = Term('leo3')
        al2 = Term('al2')
        phil = Term('phil')

        # Variables
        X = Var('X')
        Y = Var('Y')
        Z = Var('Z')

        # Clauses
        c1 = ( ancestor(X,Y) << parent(X,Y) )
        c2 = ( ancestor(X,Y) << ( parent(X,Z) & ancestor(Z,Y) ) )
        c3 = ( parent( leo3, al2 ) )
        c4 = ( parent( al2, phil ) )


..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
from __future__ import print_function
from __future__ import division  # consistent behaviour of / and // in python 2 and 3

import math
import sys


from .util import OrderedSet
from .errors import GroundingError

from collections import deque


def term2str(term):
    """Convert a term argument to string.

    :param term: the term to convert
    :type term: Term | None | int
    :return: string representation of the given term where None is converted to '_'.
    :rtype: str
    """
    if term is None:
        return '_'
    elif type(term) is int:
        if term >= 0:
            return 'A%s' % (term + 1)
        else:
            return 'X%s' % (-term)
    else:
        return str(term)


def list2term(lst):
    """Transform a Python list of terms in to a Prolog Term.

    :param lst: list of Terms
    :type lst: list of Term
    :return: Term representing a Prolog list
    :rtype: Term
    """
    from .pypl import py2pl
    tail = Term('[]')
    for e in reversed(lst):
        tail = Term('.', py2pl(e), tail)
    return tail


def term2list(term):
    """Transform a Prolog list to a Python list of terms.

    :param term: term representing a fixed length Prolog list
    :type term: Term
    :raise ValueError: given term is not a valid fixed length Prolog list
    :return: Python list containing the elements from the Prolog list
    :rtype: list of Term
    """
    result = []
    while term.functor == '.' and term.arity == 2:
        result.append(term.args[0])
        term = term.args[1]
    if not term == Term('[]'):
        raise ValueError('Expected fixed list.')
    return result


def is_ground(*terms):
    """Test whether a any of given terms contains a variable.
    :param terms: list of terms to test for the presence of variables
    :param terms: tuple of (Term | int | None)
    :return: True if none of the arguments contains any variables.
    """
    for term in terms:
        if is_variable(term):
            return False
        elif not term.is_ground():
            return False
    return True


def is_variable(term):
    """Test whether a Term represents a variable.

    :param term: term to check
    :return: True if the expression is a variable
    """
    return term is None or type(term) == int


def is_list(term):
    """Test whether a Term is a list.

    :param term: term to check
    :return: True if the term is a list.
    """
    return not is_variable(term) and term.functor == '.' and term.arity == 2


class Term(object):
    """
    A first order term, for example 'p(X,Y)'.
    :param functor: the functor of the term ('p' in the example)
    :type functor: str
    :param args: the arguments of the Term ('X' and 'Y' in the example)
    :type args: tuple of (Term | None | int)
    :param kwdargs: additional arguments; currently 'p' (probability) and 'location' \
    (character position in input)
    """

    def __init__(self, functor, *args, **kwdargs):
        self.__functor = functor
        self.__args = args
        self.__arity = len(self.__args)
        self.probability = kwdargs.get('p')
        self.location = kwdargs.get('location')
        self.op_priority = kwdargs.get('priority')
        self.op_spec = kwdargs.get('opspec')
        self.__signature = None
        self.__hash = None
        self._cache_is_ground = None
        self._cache_list_length = None
        self._cache_variables = None

    @property
    def functor(self):
        """Term functor"""
        return self.__functor

    @functor.setter
    def functor(self, value):
        """Term functor

        :param value: new value
        """
        self.__functor = value
        self.__signature = None
        self.__hash = None

    @property
    def args(self):
        """Term arguments"""
        return self.__args

    @property
    def arity(self):
        """Number of arguments"""
        return self.__arity

    @property
    def value(self):
        """Value of the Term obtained by computing the function is represents"""
        return self.compute_value()

    def compute_value(self, functions=None):
        """Compute value of the Term by computing the function it represents.

        :param functions: dictionary of user-defined functions
        :return: value of the Term
        """
        return compute_function(self.functor, self.args, functions)

    @property
    def signature(self):
        """Term's signature ``functor/arity``"""
        if self.__signature is None:
            functor = str(self.functor)
            self.__signature = '%s/%s' % (functor.strip("'"), self.arity)
        return self.__signature

    def apply(self, subst):
        """Apply the given substitution to the variables in the term.

        :param subst: A mapping from variable names to something else
        :type subst: an object with a __getitem__ method
        :raises: whatever subst.__getitem__ raises
        :returns: a new Term with all variables replaced by their values from the given substitution
        :rtype: :class:`Term`

        """
        if self.is_ground() and self.probability is None:
            # No variables to substitute.
            return self

        old_stack = [deque([self])]
        new_stack = []
        term_stack = []
        while old_stack:
            current = old_stack[-1].popleft()
            if current is None or type(current) == int:
                if new_stack:
                    new_stack[-1].append(subst[current])
                else:
                    return subst[current]
            elif current.is_var():
                if new_stack:
                    new_stack[-1].append(subst[current.name])
                else:
                    return subst[current.name]
            else:
                # Add arguments to stack
                term_stack.append(current)
                q = deque(current.args)
                if current.probability is not None:
                    q.append(current.probability)
                old_stack.append(q)
                new_stack.append([])
            while old_stack and not old_stack[-1]:
                old_stack.pop(-1)
                new_args = new_stack.pop(-1)
                term = term_stack.pop(-1)
                if term.probability is not None:
                    new_term = term.with_args(*new_args[:-1], p=new_args[-1])
                else:
                    new_term = term.with_args(*new_args)
                if new_stack:
                    new_stack[-1].append(new_term)
                else:
                    return new_term

    def __repr__(self):
        # Non-recursive version of __repr__
        stack = [deque([self])]
        # current: popleft from stack[-1]
        # arguments: new level on stack
        parts = []
        put = parts.append
        while stack:
            current = stack[-1].popleft()
            if current is None:
                put('_')
            elif type(current) == str:
                put(current)
            elif type(current) == int:
                if current < 0:
                    put('X%s' % -current)
                else:
                    put('A%s' % (current + 1))
            elif type(current) == And:  # Depends on level
                q = deque()
                q.append('(')
                if type(current.args[0]) == Or:
                    q.append('(')
                    q.append(current.args[0])
                    q.append(')')
                else:
                    q.append(current.args[0])
                tail = current.args[1]
                while isinstance(tail, Term) and tail.functor == ',' and tail.arity == 2:
                    q.append(', ')
                    if type(tail.args[0]) == Or:
                        q.append('(')
                        q.append(tail.args[0])
                        q.append(')')
                    else:
                        q.append(tail.args[0])
                    tail = tail.args[1]
                q.append(', ')
                if type(tail) == Or:
                    q.append('(')
                    q.append(tail)
                    q.append(')')
                else:
                    q.append(tail)
                q.append(')')
                stack.append(q)
            elif type(current) == Or:
                q = deque()
                q.append(current.args[0])
                tail = current.args[1]
                while isinstance(tail, Term) and tail.functor == ';' and tail.arity == 2:
                    q.append('; ')
                    q.append(tail.args[0])
                    tail = tail.args[1]
                q.append('; ')
                q.append(tail)
                stack.append(q)
            elif isinstance(current, Term) and current.functor == '.' and current.arity == 2:
                q = deque()
                q.append('[')
                q.append(current.args[0])
                tail = current.args[1]
                while isinstance(tail, Term) and tail.functor == '.' and tail.arity == 2:
                    q.append(', ')
                    q.append(tail.args[0])
                    tail = tail.args[1]
                if not tail == Term('[]'):
                    q.append(' | ')
                    q.append(tail)
                q.append(']')
                stack.append(q)
            elif isinstance(current, Term) and current.op_spec is not None:
                # Is a binary or unary operator.
                if len(current.op_spec) == 2:  # unary operator
                    put(str(current.functor).strip("'"))
                    q = deque()
                    q.append(current.args[0])
                    stack.append(q)
                else:
                    a = current.args[0]
                    b = current.args[1]
                    q = deque()
                    if not isinstance(a, Term) or a.op_priority is None or \
                            a.op_priority < current.op_priority or \
                            (a.op_priority == current.op_priority and current.op_spec == 'yfx'):
                        # no parenthesis around a
                        q.append(a)
                    else:
                        q.append('(')
                        q.append(a)
                        q.append(')')
                    op = str(current.functor)
                    if 'a' <= op[0] <= 'z':
                        q.append(' %s ' % str(current.functor).strip("'"))
                    else:
                        q.append('%s' % str(current.functor).strip("'"))
                    if not isinstance(b, Term) or \
                            b.op_priority is None or b.op_priority < current.op_priority or \
                            (b.op_priority == current.op_priority and current.op_spec == 'xfy'):
                        # no parenthesis around b
                        q.append(b)
                    else:
                        q.append('(')
                        q.append(b)
                        q.append(')')
                    stack.append(q)
            elif isinstance(current, Term):
                if current.probability is not None:
                    put(str(current.probability))  # This is a recursive call.
                    put('::')
                put(str(current.functor))
                if current.args:
                    q = deque()
                    q.append('(')
                    q.append(current.args[0])
                    for a in current.args[1:]:
                        q.append(',')
                        q.append(a)
                    q.append(')')
                    stack.append(q)
            else:
                put(str(current))
            while stack and not stack[-1]:
                stack.pop(-1)
        return ''.join(parts)

    def __call__(self, *args, **kwdargs):
        """Create a new Term with the same functor and the given arguments.

        :param args: new arguments
        :type args: tuple of (Term | None | int)
        :return:
        :rtype: Term
        """
        return self.with_args(*args, **kwdargs)

    def with_args(self, *args, **kwdargs):
        """Creates a new Term with the same functor and the given arguments.

        :param args: new arguments for the term
        :type args: tuple of (Term | int | None)
        :param kwdargs: keyword arguments for the term
        :type kwdargs: p=Constant | p=Var | p=float
        :returns: a new term with the given arguments
        :rtype: :class:`Term`

        """
        if not kwdargs and list(map(id, args)) == list(map(id, self.args)):
            return self

        if 'p' in kwdargs:
            p = kwdargs['p']
            if type(p) == float:
                p = Constant(p)
        else:
            p = self.probability

        extra = {}
        if p is not None:
            return self.__class__(self.functor, *args, p=p, location=self.location, priority=self.op_priority, opspec=self.op_spec)
        else:
            return self.__class__(self.functor, *args, location=self.location, priority=self.op_priority, opspec=self.op_spec)

    def with_probability(self, p=None):
        """Creates a new Term with the same functor and arguments but with a different probability.

        :param p: new probability (None clears the probability)
        :return: copy of the Term
        """
        return self.__class__(self.functor, *self.args, p=p, priority=self.op_priority, opspec=self.op_spec, location=self.location)

    def is_var(self):
        """Checks whether this Term represents a variable."""
        return False

    def is_constant(self):
        """Checks whether this Term represents a constant."""
        return False

    def is_ground(self):
        """Checks whether the term contains any variables."""
        if self._cache_is_ground is None:
            queue = deque([self])
            while queue:
                term = queue.popleft()
                if term is None or type(term) == int or term.is_var():
                    self._cache_is_ground = False
                    return False
                elif isinstance(term, Term):
                    if not term._cache_is_ground:
                        queue.extend(term.args)
            self._cache_is_ground = True
            return True
        else:
            return self._cache_is_ground

    def is_negated(self):
        """Checks whether the term represent a negated term."""
        return False

    def variables(self, exclude_local=False):
        """Extract the variables present in the term.

        :return: set of variables
        :rtype: :class:`problog.util.OrderedSet`
        """
        if exclude_local and self.__functor == 'findall' and self.__arity == 3:
            return self.args[2].variables()
        elif self._cache_variables is None:
            variables = OrderedSet()
            queue = deque([self])
            while queue:
                term = queue.popleft()
                if term is None or type(term) == int or term.is_var():
                    variables.add(term)
                else:
                    queue.extend(term.args)
                    if term.probability:
                        queue.append(term.probability)
            self._cache_variables = variables
        return self._cache_variables

    def _list_length(self):
        if self._cache_list_length is None:
            l = 0
            current = self
            while not is_variable(current) and current.functor == '.' and current.arity == 2:
                if current._cache_list_length is not None:
                    return l + current._cache_list_length
                l += 1
                current = current.args[1]
            self._cache_list_length = l
        return self._cache_list_length

    def _list_decompose(self):
        elements = []
        current = self
        while not is_variable(current) and current.functor == '.' and current.arity == 2:
            elements.append(current.args[0])
            current = current.args[1]
        return elements, current

    def __eq__(self, other):
        if not isinstance(other, Term):
            return False
        # Non-recursive version of equality check.
        l1 = deque([self])
        l2 = deque([other])
        while l1 and l2:
            t1 = l1.popleft()
            t2 = l2.popleft()
            if len(l1) != len(l2):
                return False
            elif id(t1) == id(t2):
                pass
            elif type(t1) != type(t2):
                return False
            elif type(t1) == int:
                if t1 != t2:
                    return False
            elif t1 is None:
                if t2 is not None:
                    return False
            elif isinstance(t1, Constant):  # t2 too
                if type(t1.functor) != type(t2.functor):
                    return False
                elif t1.functor != t2.functor:
                    return False
            else:  # t1 and t2 are Terms
                if t1.__functor != t2.__functor:
                    return False
                if t1.__arity != t2.__arity:
                    return False
                l1.extend(t1.__args)
                l2.extend(t2.__args)
        return l1 == l2  # Should both be empty.

    def _eq__list(self, other):
        """Custom equivalence test for lists.

        :param other: other Term representing a list
        :type other: Term
        :return: True if lists contain the same elements, False otherwise
        """

        if self._list_length() != other._list_length():
            return False

        elems1, tail1 = self._list_decompose()
        elems2, tail2 = other._list_decompose()

        if tail1 != tail2:
            return False
        else:
            for e1, e2 in zip(elems1, elems2):
                if e1 != e2:
                    return False
        return True

    def __hash__(self):
        if self.__hash is None:
            firstarg = None
            if len(self.__args) > 0:
                firstarg = self.__args[0]

            self.__hash = hash((self.__functor, self.__arity, firstarg, self._list_length()))
        return self.__hash

    def __lshift__(self, body):
        return Clause(self, body)

    def __and__(self, rhs):
        return And(self, rhs)

    def __or__(self, rhs):
        return Or(self, rhs)

    def __invert__(self):
        return Not('\+', self)

    def __float__(self):
        return float(self.value)

    def __int__(self):
        return int(self.value)

    def __neg__(self):
        return Not('\+', self)

    def __abs__(self):
        return self


class Var(Term):
    """A Term representing a variable.

    :param name: name of the variable
    :type name: :class:`str`

    """

    def __init__(self, name, location=None):
        Term.__init__(self, name, location=location)

    @property
    def name(self):
        """Name of the variable"""
        return self.functor

    def compute_value(self, functions=None):
        raise InstantiationError('Variables do not support evaluation: {}.'.format(self.name))

    def is_var(self):
        return True

    def is_ground(self):
        return False

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return str(other) == str(self)


class Constant(Term):
    """A constant.

        :param value: the value of the constant
        :type value: :class:`str`, :class:`float` or :class:`int`.

    """

    def __init__(self, value, location=None):
        Term.__init__(self, value, location=location)

    def compute_value(self, functions=None):
        return self.functor

    def is_constant(self):
        return True

    def __hash__(self):
        return hash(self.functor)

    def __str__(self):
        return str(self.functor)

    def is_string(self):
        """Check whether this constant is a string.

            :returns: true if the value represents a string
            :rtype: :class:`bool`
        """
        return type(self.value) == str

    def is_float(self):
        """Check whether this constant is a float.

            :returns: true if the value represents a float
            :rtype: :class:`bool`
        """
        return type(self.value) == float

    def is_integer(self):
        """Check whether this constant is an integer.

            :returns: true if the value represents an integer
            :rtype: :class:`bool`
        """
        return type(self.value) == int

    def __eq__(self, other):
        return str(self) == str(other)


class Clause(Term):
    """A clause."""

    def __init__(self, head, body, location=None):
        Term.__init__(self, ':-', head, body, location=location)
        self.head = head
        self.body = body

    def __repr__(self):
        return "%s :- %s" % (self.head, self.body)


class AnnotatedDisjunction(Term):
    """An annotated disjunction."""

    def __init__(self, heads, body, location=None):
        Term.__init__(self, ':-', heads, body, location=location)
        self.heads = heads
        self.body = body

    def __repr__(self):
        return "%s :- %s" % ('; '.join(map(str, self.heads)), self.body)


class Or(Term):
    """Or"""

    def __init__(self, op1, op2, location=None):
        Term.__init__(self, ';', op1, op2, location=location)
        self.op1 = op1
        self.op2 = op2

    @classmethod
    def from_list(cls, lst):
        """Create a disjunction based on the terms in the list.

        :param lst: list of terms
        :return: disjunction over the given terms
        """
        if lst:
            n = len(lst) - 1
            tail = lst[n]
            while n > 0:
                n -= 1
                tail = Or(lst[n], tail)
            return tail
        else:
            return Term('fail')

    def to_list(self):
        """Extract the terms of the disjunction into the list.

        :return: list of disjuncts
        """
        body = []
        current = self
        while isinstance(current, Term) and current.functor == self.functor:
            body.append(current.args[0])
            current = current.args[1]
        body.append(current)
        return body

    def __or__(self, rhs):
        return Or(self.op1, self.op2 | rhs)

    def __and__(self, rhs):
        return And(self, rhs)

    def __repr__(self):
        lhs = term2str(self.op1)
        rhs = term2str(self.op2)
        return "%s; %s" % (lhs, rhs)

    def with_args(self, *args):
        return self.__class__(*args, location=self.location)


class And(Term):
    """And"""

    def __init__(self, op1, op2, location=None):
        Term.__init__(self, ',', op1, op2, location=location)
        self.op1 = op1
        self.op2 = op2

    @classmethod
    def from_list(cls, lst):
        """Create a conjunction based on the terms in the list.

        :param lst: list of terms
        :return: conjunction over the given terms
        """
        if lst:
            n = len(lst) - 1
            tail = lst[n]
            while n > 0:
                n -= 1
                tail = And(lst[n], tail)
            return tail
        else:
            return Term('true')

    def to_list(self):
        """Extract the terms of the conjunction into the list.

        :return: list of disjuncts
        """
        body = []
        current = self
        while isinstance(current, Term) and current.functor == self.functor:
            body.append(current.args[0])
            current = current.args[1]
        body.append(current)
        return body

    def __and__(self, rhs):
        return And(self.op1, self.op2 & rhs)

    def __or__(self, rhs):
        return Or(self, rhs)

    def __repr__(self):
        lhs = term2str(self.op1)
        rhs = term2str(self.op2)
        if isinstance(self.op2, Or):
            rhs = '(%s)' % rhs
        if isinstance(self.op1, Or):
            lhs = '(%s)' % lhs

        return "%s, %s" % (lhs, rhs)

    def with_args(self, *args):
        return self.__class__(*args, location=self.location)


class Not(Term):
    """Not"""

    def __init__(self, functor, child, location=None, **kwdargs):
        Term.__init__(self, functor, child, location=location)
        self.child = child

    def __repr__(self):
        c = str(self.child)
        if isinstance(self.child, And) or isinstance(self.child, Or):
            c = '(%s)' % c
        if self.functor == 'not':
            return 'not %s' % c
        else:
            return '%s%s' % (self.functor, c)

    def is_negated(self):
        return True

    def __neg__(self):
        return self.child

    def __abs__(self):
        return -self


_arithmetic_functions = {
    ("+", 2): (lambda a, b: a + b),
    ("-", 2): (lambda a, b: a - b),
    ("/\\", 2): (lambda a, b: a & b),
    ("\\/", 2): (lambda a, b: a | b),
    ("xor", 2): (lambda a, b: a ^ b),
    ("xor", 2): (lambda a, b: a ^ b),
    ("#", 2): (lambda a, b: a ^ b),
    ("><", 2): (lambda a, b: a ^ b),
    ("*", 2): (lambda a, b: a * b),
    ("/", 2): (lambda a, b: a / b),
    ("//", 2): (lambda a, b: a // b),
    ("<<", 2): (lambda a, b: a << b),
    (">>", 2): (lambda a, b: a >> b),
    ("mod", 2): (lambda a, b: a % b),
    ("mod", 2): (lambda a, b: a % b),
    ("rem", 2): (lambda a, b: a % b),
    ("rem", 2): (lambda a, b: a % b),
    ("div", 2): (lambda a, b: (a - (a % b)) // b),
    ("div", 2): (lambda a, b: (a - (a % b)) // b),
    ("**", 2): (lambda a, b: a ** b),
    ("^", 2): (lambda a, b: a ** b),
    ("+", 1): (lambda a: a),
    ("-", 1): (lambda a: -a),
    ("\\", 1): (lambda a: ~a),
    ("atan", 2): math.atan2,
    ("atan2", 2): math.atan2,
    ("integer", 1): int,
    ("float", 1): float,
    ("float_integer_part", 1): lambda f: int(f),
    ("float_fractional_part", 1): lambda f: f - int(f),
    ("abs", 1): abs,
    ("ceiling", 1): lambda x: int(math.ceil(x)),
    ("round", 1): lambda x: int(round(x)),
    ("floor", 1): lambda x: int(math.floor(x)),
    ("truncate", 1): lambda x: int(math.trunc),
    ("min", 2): min,
    ("max", 2): max,
    ("exp", 2): math.pow,
    ("epsilon", 0): lambda: sys.float_info.epsilon,
    ("inf", 0): lambda: float('inf'),
    ("nan", 0): lambda: float('nan'),
    ("sign", 1): lambda x: 1 if x > 0 else -1 if x < 0 else 0

}

_from_math_1 = ['exp', 'log', 'log10', 'sqrt', 'sin', 'cos', 'tan', 'asin', 'acos', 'atan',
                'sinh', 'cosh', 'tanh', 'asinh', 'acosh', 'atanh', 'lgamma', 'gamma', 'erf',
                'erfc']
for _f in _from_math_1:
    _arithmetic_functions[(_f, 1)] = getattr(math, _f)

_from_math_0 = ['pi', 'e']
for _f in _from_math_0:
    _arithmetic_functions[(_f, 0)] = lambda: getattr(math, _f)


def unquote(s):
    """Strip single quotes from the string.

    :param s: string to remove quotes from
    :return: string with quotes removed
    """
    return s.strip("'")


def compute_function(func, args, extra_functions=None):
    """Compute the result of an arithmetic function given by a functor and a list of arguments.

    :param func: functor
    :type: basestring
    :param args: arguments
    :type args: (list | tuple) of (Term | int | None)
    :param extra_functions: additional user-defined functions
    :raises: ArithmeticError if the function is unknown or if an error occurs while computing it
    :return: result of the function
    :rtype: Constant
    """
    if extra_functions is None:
        extra_functions = {}

    function = _arithmetic_functions.get((unquote(func), len(args)))
    if function is None:
        function = extra_functions.get((unquote(func), len(args)))
        if function is None:
            raise ArithmeticError("Unknown function '%s'/%s" % (func, len(args)))
    try:
        values = [arg.compute_value(extra_functions) for arg in args]
        if None in values:
            return None
        else:
            return function(*values)
    except ZeroDivisionError:
        raise ArithmeticError("Division by zero.")


class InstantiationError(GroundingError):
    """Error used when performing arithmetic with a non-ground term."""
    pass


# noinspection PyShadowingBuiltins
class ArithmeticError(GroundingError):
    """Error used when an error occurs during evaluation of an arithmetic expression."""
    pass


#! /usr/bin/env python

"""
Learning from interpretations
-----------------------------

Parameter learning for ProbLog.

Given a probabilistic program with parameterized weights and a set of partial implementations,
learns appropriate values of the parameters.

Algorithm
+++++++++

The algorithm operates as follows:

    0. Set initial values for the weights to learn.
    1. Set the evidence present in the example.
    2. Query the model for the weights of the atoms to be learned.
    3. Update the weights to learn by taking the mean value over all examples and queries.
    4. Repeat steps 1 to 4 until convergence (or a maximum number of iterations).
    
The score of the model for a given example is obtained by calculating the probability of the
evidence in the example.

Implementation
++++++++++++++

The algorithm is implemented on top of the ProbLog toolbox.

It uses the following extensions of ProbLog's classes:

    * a LogicProgram implementation that rewrites the model and extracts the weights to learn
    (see :py:func:`learning.lfi.LFIProblem.__iter__`)
    * a custom semiring that looks up the current value of a weight to learn
    (see :py:func:`learning.lfi.LFIProblem.value`)


.. autoclass:: learning.lfi.LFIProblem
    :members: __iter__, value

"""

from __future__ import print_function

import sys
import os
import random
import math
import logging

from collections import defaultdict

# Make sure the ProbLog module is on the path.
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../')))

from problog.engine import DefaultEngine, ground
from problog.evaluator import SemiringProbability
from problog.logic import Term, Constant, Clause, AnnotatedDisjunction, Or, Var
from problog.program import PrologString, PrologFile, LogicProgram
from problog.core import ProbLogError
from problog.errors import process_error, InconsistentEvidenceError
from problog.sdd_formula import SDD

from problog import get_evaluatable, get_evaluatables
import traceback


def str2bool(s):
    if str(s) == 'true':
        return True
    elif str(s) == 'false':
        return False
    else:
        return None


class LFIProblem(SemiringProbability, LogicProgram) :
    
    def __init__(self, source, examples, max_iter=10000, min_improv=1e-10, verbose=0, knowledge=SDD, leakprob=None, **extra):
        """
        :param source:
        :param examples:
        :param max_iter:
        :param min_improv:
        :param verbose:
        :param knowledge:
        :param leakprob: Add all true evidence atoms with the given probability
                         to avoid 'inconsistent evidence' errors. This also
                         allows to learn a program without constants and
                         retrieve the constants from the evidence file.
                         (default: None)
        :param extra:
        :return:
        """
        SemiringProbability.__init__(self)
        LogicProgram.__init__(self)
        self.source = source
        self.names = []
        self.queries = []
        self.weights = []
        self.examples = examples
        self.leakprob = leakprob
        self.leakprobatoms = None
        self._compiled_examples = None
        
        self.max_iter = max_iter
        self.min_improv = min_improv
        self.verbose = verbose
        self.iteration = 0

        self.knowledge = knowledge

        self.output_mode = False
    
    def value(self, a):
        """Overrides from SemiringProbability.
        Replaces weights of the form ``lfi(i)`` by their current estimated value.
        """
        
        if isinstance(a, Term) and a.functor == 'lfi':
            assert(len(a.args) == 1)
            index = int(a.args[0])
            return self.weights[index]
        else :
            return float(a)
         
    @property 
    def count(self):
        """Number of parameters to learn."""
        return len(self.weights)
    
    def prepare(self):
        """Prepare for learning."""
        self._compile_examples()
        
    def _process_examples(self):
        """Process examples by grouping together examples with similar structure.
    
        :return: example groups based on evidence atoms
        :rtype: dict of atoms : values for examples
        """
    
        # value can be True / False / None
        # ( atom ), ( ( value, ... ), ... ) 

        # Simple implementation: don't add neutral evidence.
        result = defaultdict(list)
        for example in self.examples:
            atoms, values = zip(*example)
            result[atoms].append(values)
        return result
    
    def _compile_examples(self):
        """Compile examples.
    
        :param examples: Output of ::func::`process_examples`.
        """
        logger = logging.getLogger('problog_lfi')

        baseprogram = DefaultEngine().prepare(self)
        examples = self._process_examples()
        result = []
        for atoms, example_group in examples.items():
            ground_program = None   # Let the grounder decide
            for n, example in enumerate(example_group):
                if self.verbose:
                    logger.debug('Compiling example %s ...' % n)

                ground_program = ground(baseprogram, ground_program,
                                        evidence=list(zip(atoms, example)))
                for i, node, t in ground_program:
                    if t == 'atom' and \
                            isinstance(node.probability, Term) and \
                            node.probability.functor == 'lfi':
                        factname = 'lfi_fact_%s' % node.probability.args[0]
                        factargs = ()
                        if type(node.identifier) == tuple:
                             factargs = node.identifier[1]
                        fact = Term(factname, *factargs)
                        ground_program.add_query(fact, i)
                compiled_program = self.knowledge.create_from(ground_program)
                result.append((atoms, example, compiled_program, n))
        self._compiled_examples = result

    def _process_atom(self, atom, body):
        """Returns tuple ( prob_atom, [ additional clauses ] )"""
        if isinstance(atom, Or):
            # Annotated disjunction
            atoms = atom.to_list()
        else:
            atoms = [atom]

        atoms_out = []
        extra_clauses = []

        has_lfi_fact = False
        available_probability = 1.0

        num_random_weights = 0
        for atom in atoms:
            if atom.probability and atom.probability.functor == 't':
                start_value = atom.probability.args[0]
                if start_value.is_var():
                    num_random_weights += 1
                else:
                    available_probability -= float(start_value)
            elif atom.probability and atom.is_constant():
                available_probability -= float(atom.probability)

        random_weights = [random.random() for i in range(0, num_random_weights+1)]
        norm_factor = available_probability / sum(random_weights)
        random_weights = [r*norm_factor for r in random_weights]

        for atom in atoms:
            if atom.probability and atom.probability.functor == 't':
                # t(_)::p(X) :- body.
                #
                # Translate to
                #   lfi(1)::lfi_fact_1(X).
                #   p(X) :- body, lfi_fact1(X).
                # For annotated disjunction: t(_)::p1(X); t(_)::p2(X) :- body.
                #   lfi1::lfi_fact1(X); lfi2::lfi_fact2(X); ... .
                #   p1(X) :- body, lfi_fact1(X).
                #   p2(X) :- body, lfi_fact2(X).
                #  ....
                has_lfi_fact = True

                # Learnable probability
                assert(len(atom.probability.args) == 1)
                start_value = atom.probability.args[0]

                # Replace anonymous variables with non-anonymous variables.
                class ReplaceAnon(object):

                    def __init__(self):
                        self.cnt = 0

                    def __getitem__(self, key):
                        if key == '_':
                            self.cnt += 1
                            return Var('anon_%s' % self.cnt)
                        else:
                            return Var(key)

                atom1 = atom.apply(ReplaceAnon())

                # 1) Introduce a new fact
                lfi_fact = Term('lfi_fact_%d' % self.count, *atom1.args)
                lfi_prob = Term('lfi', Constant(self.count))

                # 2) Replacement atom
                replacement = lfi_fact.with_probability(lfi_prob)
                if body is None:
                    new_body = lfi_fact
                else:
                    new_body = body & lfi_fact

                # 3) Create redirection clause
                extra_clauses += [Clause(atom1.with_probability(), new_body)]

                # 4) Set initial weight
                if start_value.is_var():
                    self.weights.append(random_weights.pop(-1))
                else:
                    self.weights.append(float(start_value))

                # # 5) Add query
                # self.queries.append(lfi_fact)
                # if body:
                #     extra_clauses.append(Clause(Term('query', lfi_fact), body))
                # else:
                #     extra_clauses.append(Term('query', lfi_fact))

                # 6) Add name
                self.names.append(atom)

                atoms_out.append(replacement)
            else:
                atoms_out.append(atom)

        if has_lfi_fact:
            if len(atoms) == 1:     # Simple clause
                return [atoms_out[0]] + extra_clauses
            else:
                return [AnnotatedDisjunction(atoms_out, Term('true'))] + extra_clauses
        else:
            if len(atoms) == 1:
                if body is None:
                    return [atoms_out[0]]
                else:
                    return [Clause(atoms_out[0], body)]
            else:
                if body is None:
                    body = Term('true')
                return [AnnotatedDisjunction(atoms_out, body)]

    def _process_atom_output(self, atom, body):
        """Returns tuple ( prob_atom, [ additional clauses ] )"""

        if isinstance(atom, Or):
            atoms = atom.to_list()
        else:
            atoms = [atom]

        atoms_out = []
        for atom in atoms:
            if atom.probability and atom.probability.functor == 't':
                assert (atom in self.names)
                index = self.output_names.index(atom)
                self.output_names[index] = None
                weight = self.weights[index]
                result = atom.with_probability(weight)
                atoms_out.append(result)
            else:
                atoms_out.append(atom)
        if len(atoms_out) == 1:
            if body is None:
                return [atoms_out[0]]
            else:
                return [Clause(atoms_out[0], body)]
        else:
            return [AnnotatedDisjunction(atoms_out, body)]
        
    # Overwrite from LogicProgram    
    def __iter__(self) :
        """
        Iterate over the clauses of the source model.
        This object can be used as a LogicProgram to be passed to the grounding Engine.
        
        Extracts and processes all ``t(...)`` weights.
        This
        
            * replaces each probabilistic atom ``t(...)::p(X)`` by a unique atom \
            ``lfi(i) :: lfi_fact_i(X)``;
            * adds a new clause ``p(X) :- lfi_fact_i(X)``;
            * adds a new query ``query( lfi_fact_i(X) )``;
            * initializes the weight of ``lfi(i)`` based on the ``t(...)`` specification;
        
        This also removes all existing queries from the model.
        
        Example:
        
        .. code-block:: prolog
        
            t(_) :: p(X) :- b(X).
            t(_) :: p(X) :- c(X).

        is transformed into
        
        .. code-block:: prolog
        
            lfi(0) :: lfi_fact_0(X) :- b(X).
            p(X) :- lfi_fact_0(X).
            lfi(1) :: lfi_fact_1(X) :- c(X).
            p(X) :- lfi_fact_1(X).
            query(lfi_fact_0(X)).
            query(lfi_fact_1(X)).

        If ``self.leakprobs`` is a value, then during learning all true
        examples are added to the program with the given leak probability.

        """

        if self.output_mode:
            process_atom = self._process_atom_output
            self.output_names = self.names[:]
        else:
            process_atom = self._process_atom

        for clause in self.source:
            if isinstance(clause, Clause):
                if clause.head.functor == 'query' and clause.head.arity == 1:
                    continue
                extra_clauses = process_atom(clause.head, clause.body)
                for extra in extra_clauses:
                    yield extra
            elif isinstance(clause, AnnotatedDisjunction):
                extra_clauses = process_atom(Or.from_list(clause.heads), clause.body)
                for extra in extra_clauses:
                    yield extra
            else:
                if clause.functor == 'query' and clause.arity == 1:
                    continue
                # Fact
                extra_clauses = process_atom(clause, None)
                for extra in extra_clauses:
                    yield extra

        if self.leakprob is not None:
            leakprob_atoms = self._get_leakprobatoms()
            for example_atom in leakprob_atoms:
                yield example_atom.with_probability(Constant(self.leakprob))

    def _get_leakprobatoms(self):
        if self.leakprobatoms is not None:
            return self.leakprobatoms
        self.leakprobatoms = set()
        for examples in self.examples:
            for example, obs in examples:
                if obs:
                    self.leakprobatoms.add(example)
        return self.leakprobatoms

    def _evaluate_examples( self ) :
        """Evaluate the model with its current estimates for all examples."""
        
        results = []
        i = 0
        logging.getLogger('problog_lfi').debug('Evaluating examples ...')
        for at, val, comp, n in self._compiled_examples:
            evidence = {}
            for a, v in zip(at, map(str2bool, val)):
                if a in evidence:
                    if evidence[a] != v:
                        context = ' (found evidence({},{}) and evidence({},{}) in example {})'.format(a, evidence[a], a, v, n+1)
                        raise InconsistentEvidenceError(source=a, context=context)
                else:
                    evidence[a] = v
            try:
                evaluator = comp.get_evaluator(semiring=self, evidence=evidence)
            except InconsistentEvidenceError as err:
                if err.context == '':
                    context = ' (example {})'.format(n+1)
                else:
                    context = err.context + ' (example {})'.format(n+1)
                raise InconsistentEvidenceError(err.source, context)
            p_queries = {}
            # Probability of query given evidence
            for name, node in evaluator.formula.queries():
                w = evaluator.evaluate_fact(node)
                if w < 1e-6:
                    p_queries[name] = 0.0
                else:
                    p_queries[name] = w
            p_evidence = evaluator.evaluate_evidence()
            i += 1
            results.append((p_evidence, p_queries))
        return results
    
    def _update(self, results) :
        """Update the current estimates based on the latest evaluation results."""
        
        fact_marg = [0.0] * self.count
        fact_count = [0] * self.count
        score = 0.0
        for pEvidence, result in results:
            for fact, value in result.items():
                fact = str(fact)
                index = int(fact.split('(')[0].rsplit('_',1)[1])
                fact_marg[index] += value
                fact_count[index] += 1
            try:
                score += math.log(pEvidence)
            except ValueError:
                raise ProbLogError('Inconsistent evidence.')

        output = {}
        for index in range(0, self.count) :
            if fact_count[index] > 0 :
                self.weights[index] = fact_marg[index] / fact_count[index]
        return score
        
    def step(self) :
        self.iteration += 1
        results = self._evaluate_examples()
        return self._update(results)

    def get_model(self):
        self.output_mode = True
        lines = []
        for l in self:
            lines.append('%s.' % l)
        lines.append('')
        self.output_mode = False
        return '\n'.join(lines)
        
    def run(self) :
        self.prepare()
        logging.getLogger('problog_lfi').info('Weights to learn: %s' % self.names)
        logging.getLogger('problog_lfi').info('Initial weights: %s' % self.weights)
        delta = 1000
        prev_score = -1e10
        while self.iteration < self.max_iter and (delta < 0 or delta > self.min_improv):
            score = self.step()
            logging.getLogger('problog_lfi').info('Weights after iteration %s: %s' % (self.iteration, self.weights))
            delta = score - prev_score
            prev_score = score
        return prev_score

def extract_evidence(pl):
    engine = DefaultEngine()
    atoms = engine.query(pl, Term('evidence', None, None))
    atoms1 = engine.query(pl, Term('evidence', None))
    for atom in atoms1:
        atom = atom[0]
        if atom.is_negated():
            atoms.append((-atom, Term('false')))
        else:
            atoms.append((atom, Term('true')))
    return atoms

def read_examples(*filenames):
    
    for filename in filenames:
        engine = DefaultEngine()
        
        with open(filename) as f:
            example = ''
            for line in f:
                if line.strip().startswith('---') :
                    pl = PrologString(example)
                    atoms = extract_evidence(pl)
                    if len(atoms) > 0:
                        yield atoms
                    example = ''
                else:
                    example += line
            if example:
                pl = PrologString(example)
                atoms = extract_evidence(pl)
                if len(atoms) > 0:
                    yield atoms
    
    
def run_lfi( program, examples, output_model=None, **kwdargs):
    lfi = LFIProblem(program, examples, **kwdargs)
    score = lfi.run()

    if output_model is not None:
        with open(output_model, 'w') as f:
            f.write(lfi.get_model())
    return score, lfi.weights, lfi.names, lfi.iteration


def argparser():
    import argparse
    parser = argparse.ArgumentParser(description="Learning from interpretations with ProbLog")
    parser.add_argument('model')
    parser.add_argument('examples', nargs='+')
    parser.add_argument('-n', dest='max_iter', default=10000, type=int )
    parser.add_argument('-d', dest='min_improv', default=1e-10, type=float )
    parser.add_argument('-O', '--output-model', type=str, default=None,
                        help='write resulting model to given file')
    parser.add_argument('-o', '--output', type=str, default=None,
                        help='write output to file')
    parser.add_argument('-k', '--knowledge', dest='koption', choices=get_evaluatables(),
                        default=None, help='knowledge compilation tool')
    parser.add_argument('-l', '--leak-probabilities', dest='leakprob', type=float,
                        help='Add leak probabilities for evidence atoms.')
    parser.add_argument('-v', '--verbose', action='count', default=0)
    parser.add_argument('--web', action='store_true', help=argparse.SUPPRESS)
    return parser


def create_logger(name, verbose):
    levels = [logging.WARNING, logging.INFO, logging.DEBUG] + list(range(9, 0, -1))
    verbose = max(0, min(len(levels)-1, verbose))
    logger = logging.getLogger(name)
    ch = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('[%(levelname)s] %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.setLevel(levels[verbose])


def main(argv, result_handler=None):
    parser = argparser()
    args = parser.parse_args(argv)

    if result_handler is None:
        if args.web:
            result_handler = print_result_json
        else:
            result_handler = print_result

    knowledge = get_evaluatable(args.koption)

    if args.output is None:
        outf = sys.stdout
    else:
        outf = open(args.output, 'w')

    create_logger('problog_lfi', args.verbose)
    create_logger('problog', args.verbose - 1)

    program = PrologFile(args.model)
    examples = list(read_examples(*args.examples))
    if len(examples) == 0:
        logging.getLogger('problog_lfi').warn('no examples specified')
    else:
        logging.getLogger('problog_lfi').info('Number of examples: %s' % len(examples))
    options = vars(args)
    del options['examples']

    try:
        results = run_lfi(program, examples, knowledge=knowledge, **options)

        for n in results[2]:
            n.loc = program.lineno(n.location)
        retcode = result_handler((True, results), output=outf)
    except Exception as err:
        trace = traceback.format_exc()
        err.trace = trace
        retcode = result_handler((False, err), output=outf)

    if args.output is not None:
        outf.close()

    if retcode:
        sys.exit(retcode)


def print_result(d, output, precision=8):
    success, d = d
    if success:
        score, weights, names, iterations = d
        weights = list(map(lambda x: round(x, precision), weights))
        print (score, weights, names, iterations, file=output)
        return 0
    else:
        print (process_error(d), file=output)
        return 1


def print_result_json(d, output, precision=8):
    import json
    success, d = d
    if success:
        score, weights, names, iterations = d
        results = {'SUCCESS': True,
                   'score': score,
                   'iterations': iterations,
                   'weights': [[str(n.with_probability()), round(w, precision), n.loc[1], n.loc[2]]
                               for n, w in zip(names, weights)],
                   }
        print (json.dumps(results), file=output)
    else:
        results = {'SUCCESS': False, 'err': vars(d)}
        print (json.dumps(results), file=output)
    return 0


if __name__ == '__main__':
    main(sys.argv[1:])



"""
problog.evaluator - Commone interface for evaluation
----------------------------------------------------

Provides common interface for evaluation of weighted logic formulas.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from __future__ import print_function

import math

from .core import ProbLogObject, transform_allow_subclass
from .errors import InconsistentEvidenceError, InvalidValue


class OperationNotSupported(Exception):

    def __init__(self):
        Exception.__init__(self, 'This operation is not supported by this semiring.')


class Semiring(object):
    """Interface for weight manipulation.

    A semiring is a set R equipped with two binary operations '+' and 'x'.

    The semiring can use different representations for internal values and external values.
    For example, the LogProbability semiring uses probabilities [0, 1] as external values and uses \
     the logarithm of these probabilities as internal values.

    Most methods take and return internal values. The execeptions are:

       - value, pos_value, neg_value: transform an external value to an internal value
       - result: transform an internal to an external value
       - result_zero, result_one: return an external value

    """

    def one(self):
        """Returns the identity element of the multiplication."""
        raise NotImplementedError()

    def is_one(self, value):
        """Tests whether the given value is the identity element of the multiplication."""
        return value == self.one

    def zero(self):
        """Returns the identity element of the addition."""
        raise NotImplementedError()

    def is_zero(self, value):
        """Tests whether the given value is the identity element of the addition."""
        return value == self.zero

    def plus(self, a, b):
        """Computes the addition of the given values."""
        raise NotImplementedError()

    def times(self, a, b):
        """Computes the multiplication of the given values."""
        raise NotImplementedError()

    def negate(self, a):
        """Returns the negation. This operation is optional.
        For example, for probabilities return 1-a.

        :raise OperationNotSupported: if the semiring does not support this operation
        """
        raise OperationNotSupported()

    def value(self, a):
        """Transform the given external value into an internal value."""
        return float(a)

    def result(self, a, formula=None):
        """Transform the given internal value into an external value."""
        return a

    def normalize(self, a, z):
        """Normalizes the given value with the given normalization constant.

        For example, for probabilities, returns a/z.

        :raise OperationNotSupported: if z is not one and the semiring does not support \
         this operation
        """
        if self.is_one(z):
            return a
        else:
            raise OperationNotSupported()

    def pos_value(self, a, key=None):
        """Extract the positive internal value for the given external value."""
        return self.value(a)

    def neg_value(self, a, key=None):
        """Extract the negative internal value for the given external value."""
        return self.negate(self.value(a))

    def result_zero(self):
        """Give the external representation of the identity element of the addition."""
        return self.result(self.zero())

    def result_one(self):
        """Give the external representation of the identity element of the multiplication."""
        return self.result(self.one())

    def is_dsp(self):
        """Indicates whether this semiring requires solving a disjoint sum problem."""
        return False

    def is_nsp(self):
        """Indicates whether this semiring requires solving a neutral sum problem."""
        return False

    def in_domain(self, a):
        """Checks whether the given (internal) value is valid."""
        return True


class SemiringProbability(Semiring):
    """Implementation of the semiring interface for probabilities."""

    def one(self):
        return 1.0

    def zero(self):
        return 0.0

    def is_one(self, value):
        return 1.0 - 1e-12 < value < 1.0 + 1e-12

    def is_zero(self, value):
        return -1e-12 < value < 1e-12

    def plus(self, a, b):
        return a + b

    def times(self, a, b):
        return a * b

    def negate(self, a):
        return 1.0 - a

    def normalize(self, a, z):
        return a / z

    def value(self, a):
        v = float(a)
        if 0.0 - 1e-9 <= v <= 1.0 + 1e-9:
            return v
        else:
            raise InvalidValue("Not a valid value for this semiring: '%s'" % a, location=a.location)

    def is_dsp(self):
        """Indicates whether this semiring requires solving a disjoint sum problem."""
        return True

    def in_domain(self, a):
        return 0.0 - 1e-9 <= a <= 1.0 + 1e-9


class SemiringLogProbability(SemiringProbability):
    """Implementation of the semiring interface for probabilities with logspace calculations."""

    inf, ninf = float("inf"), float("-inf")

    def one(self):
        return 0.0

    def zero(self):
        return self.ninf

    def is_zero(self, value):
        return value <= -1e100

    def is_one(self, value):
        return -1e-12 < value < 1e-12

    def plus(self, a, b):
        if a < b:
            if a == self.ninf:
                return b
            return b + math.log1p(math.exp(a - b))
        else:
            if b == self.ninf:
                return a
            return a + math.log1p(math.exp(b - a))

    def times(self, a, b):
        return a + b

    def negate(self, a):
        if a > -1e-10:
            return self.zero()
        return math.log1p(-math.exp(a))

    def value(self, a):
        v = float(a)
        if -1e-9 <= v < 1e-9:
            return self.zero()
        else:
            if 0.0 - 1e-9 <= v <= 1.0 + 1e-9:
                return math.log(v)
            else:
                raise InvalidValue("Not a valid value for this semiring: '%s'" % a, location=a.location)

    def result(self, a, formula=None):
        return math.exp(a)

    def normalize(self, a, z):
        # Assumes Z is in log
        return a - z

    def is_dsp(self):
        """Indicates whether this semiring requires solving a disjoint sum problem."""
        return True

    def in_domain(self, a):
        return a <= 1e-12


class SemiringSymbolic(Semiring):
    """Implementation of the semiring interface for probabilities using symbolic calculations."""

    def one(self):
        return "1"

    def zero(self):
        return "0"

    def plus(self, a, b):
        if a == "0":
            return b
        elif b == "0":
            return a
        else:
            return "(%s + %s)" % (a, b)

    def times(self, a, b):
        if a == "0" or b == "0":
            return "0"
        elif a == "1":
            return b
        elif b == "1":
            return a
        else:
            return "%s*%s" % (a, b)

    def negate(self, a):
        if a == "0":
            return "1"
        elif a == "1":
            return "0"
        else:
            return "(1-%s)" % a

    def value(self, a):
        return str(a)

    def normalize(self, a, z):
        if z == "1":
            return a
        else:
            return "%s / %s" % (a, z)

    def is_dsp(self):
        """Indicates whether this semiring requires solving a disjoint sum problem."""
        return True



@transform_allow_subclass
class Evaluatable(ProbLogObject):
    """Interface for evaluatable formulae."""

    def evidence_all(self):
        raise NotImplementedError()

    def _create_evaluator(self, semiring, weights, **kwargs):
        """Create a new evaluator.

        :param semiring: semiring to use
        :param weights: weights to use (replace weights defined in formula)
        :return: evaluator
        :rtype: Evaluator
        """
        raise NotImplementedError('Evaluatable._create_evaluator is an abstract method')

    def get_evaluator(self, semiring=None, evidence=None, weights=None, **kwargs):
        """Get an evaluator for computing queries on this formula.
        It creates an new evaluator and initializes it with the given or predefined evidence.

        :param semiring: semiring to use
        :param evidence: evidence values (override values defined in formula)
        :type evidence: dict(Term, bool)
        :param weights: weights to use
        :return: evaluator for this formula
        """
        if semiring is None:
            semiring = SemiringLogProbability()

        evaluator = self._create_evaluator(semiring, weights, **kwargs)

        for ev_name, ev_index, ev_value in self.evidence_all():
            if ev_index == 0 and ev_value > 0:
                pass  # true evidence is deterministically true
            elif ev_index is None and ev_value < 0:
                pass  # false evidence is deterministically false
            elif ev_index == 0 and ev_value < 0:
                raise InconsistentEvidenceError(source='evidence('+str(ev_name)+',false)')  # true evidence is false
            elif ev_index is None and ev_value > 0:
                raise InconsistentEvidenceError(source='evidence('+str(ev_name)+',true)')  # false evidence is true
            elif evidence is None and ev_value != 0:
                evaluator.add_evidence(ev_value * ev_index)
            else:
                try:
                    value = evidence[ev_name]
                    if value:
                        evaluator.add_evidence(ev_index)
                    else:
                        evaluator.add_evidence(-ev_index)
                except KeyError:
                    pass

        evaluator.propagate()
        return evaluator

    def evaluate(self, index=None, semiring=None, evidence=None, weights=None, **kwargs):
        """Evaluate a set of nodes.

        :param index: node to evaluate (default: all queries)
        :param semiring: use the given semiring
        :param evidence: use the given evidence values (overrides formula)
        :param weights: use the given weights (overrides formula)
        :return: The result of the evaluation expressed as an external value of the semiring. \
         If index is ``None`` (all queries) then the result is a dictionary of name to value.
        """
        evaluator = self.get_evaluator(semiring, evidence, weights, **kwargs)

        if index is None:
            result = {}
            # Probability of query given evidence

            # interrupted = False
            for name, node in evaluator.formula.queries():
                w = evaluator.evaluate(node)
                result[name] = w
            return result
        else:
            return evaluator.evaluate(index)


class Evaluator(object):
    """Generic evaluator."""

    # noinspection PyUnusedLocal
    def __init__(self, formula, semiring, weights, **kwargs):
        self.formula = formula
        self.weights = {}
        self.given_weights = weights

        self.__semiring = semiring

        self.__evidence = []

    @property
    def semiring(self):
        """Semiring used by this evaluator."""
        return self.__semiring

    def propagate(self):
        """Propagate changes in weight or evidence values."""
        raise NotImplementedError('Evaluator.propagate() is an abstract method.')

    def evaluate(self, index):
        """Compute the value of the given node."""
        raise NotImplementedError('abstract method')

    def evaluate_evidence(self):
        raise NotImplementedError('abstract method')

    def evaluate_fact(self, node):
        """Evaluate fact.

        :param node: fact to evaluate
        :return: weight of the fact (as semiring result value)
        """
        raise NotImplementedError('abstract method')

    def add_evidence(self, node):
        """Add evidence"""
        self.__evidence.append(node)

    def has_evidence(self):
        """Checks whether there is active evidence."""
        return self.__evidence != []

    def set_evidence(self, index, value):
        """Set value for evidence node.

        :param index: index of evidence node
        :param value: value of evidence
        """
        raise NotImplementedError('abstract method')

    def set_weight(self, index, pos, neg):
        """Set weight of a node.

        :param index: index of node
        :param pos: positive weight (as semiring internal value)
        :param neg: negative weight (as semiring internal value)
        """
        raise NotImplementedError('abstract method')

    def clear_evidence(self):
        """Clear all evidence."""
        self.__evidence = []

    def evidence(self):
        """Iterate over evidence."""
        return iter(self.__evidence)


class FormulaEvaluator(object):
    """Standard evaluator for boolean formula."""

    def __init__(self, formula, semiring):
        self._computed_weights = {}
        self._semiring = semiring
        self._formula = formula
        self._fact_weights = {}

    @property
    def semiring(self):
        return self._semiring

    @property
    def formula(self):
        return self._formula

    def set_weights(self, weights):
        """Set known weights.

        :param weights: dictionary of weights
        :return:
        """
        self._computed_weights.clear()
        self._fact_weights = weights

    def get_weight(self, index):
        """Get the weight of the node with the given index.

        :param index: integer or formula.TRUE or formula.FALSE
        :return: weight of the node
        """

        if index == self.formula.TRUE:
            return self.semiring.one()
        elif index == self.formula.FALSE:
            return self.semiring.zero()
        elif index < 0:
            weight = self._fact_weights.get(index)
            if weight is None:
                # This will only work if the semiring support negation!
                nw = self.get_weight(-index)
                return self.semiring.negate(nw)
            else:
                return weight[1]
        else:
            weight = self._fact_weights.get(index)
            if weight is None:
                weight = self._computed_weights.get(index)
                if weight is None:
                    weight = self.compute_weight(index)
                    self._computed_weights[index] = weight
                return weight
            else:
                return weight[0]

    def evaluate(self, index):
        return self.semiring.result(self.get_weight(index), self.formula)

    def compute_weight(self, index):
        """Compute the weight of the node with the given index.

        :param index: integer or formula.TRUE or formula.FALSE
        :return: weight of the node
        """

        if index == self.formula.TRUE:
            return self.semiring.one()
        elif index == self.formula.FALSE:
            return self.semiring.zero()
        else:
            node = self.formula.get_node(abs(index))
            ntype = type(node).__name__

            if ntype == 'atom':
                return self.semiring.one()
            else:
                childprobs = [self.get_weight(c) for c in node.children]
                if ntype == 'conj':
                    p = self.semiring.one()
                    for c in childprobs:
                        p = self.semiring.times(p, c)
                    return p
                elif ntype == 'disj':
                    p = self.semiring.zero()
                    for c in childprobs:
                        p = self.semiring.plus(p, c)
                    return p
                else:
                    raise TypeError("Unexpected node type: '%s'." % ntype)


class FormulaEvaluatorNSP(FormulaEvaluator):
    """Evaluator for boolean formula that addresses the Neutral Sum Problem."""

    def __init__(self, formula, semiring):
        FormulaEvaluator.__init__(self, formula, semiring)

    def get_weight(self, index):
        """Get the weight of the node with the given index.

        :param index: integer or formula.TRUE or formula.FALSE
        :return: weight of the node
        """

        if index == self.formula.TRUE:
            return self.semiring.one()
        elif index == self.formula.FALSE:
            return self.semiring.zero()
        elif index < 0:
            weight = self._fact_weights.get(-index)
            if weight is None:
                # This will only work if the semiring support negation!
                nw, nu = self.get_weight(-index)
                return self.semiring.negate(nw), nu
            else:
                return weight[1], {abs(index)}
        else:
            weight = self._fact_weights.get(index)
            if weight is None:
                weight = self._computed_weights.get(index)
                if weight is None:
                    weight = self.compute_weight(index)
                    self._computed_weights[index] = weight
                return weight
            else:
                return weight[0], {abs(index)}

    def evaluate(self, index):
        cp, cu = self.get_weight(index)
        all_used = set(self._fact_weights.keys())

        not_used = all_used - cu
        for nu in not_used:
            nu_p, a = self.get_weight(nu)
            nu_n, b = self.get_weight(-nu)
            cp = self.semiring.times(cp, self.semiring.plus(nu_p, nu_n))

        return self.semiring.result(cp, self.formula)

    def compute_weight(self, index):
        """Compute the weight of the node with the given index.

        :param index: integer or formula.TRUE or formula.FALSE
        :return: weight of the node
        """

        node = self.formula.get_node(index)
        ntype = type(node).__name__

        if ntype == 'atom':
            return self.semiring.one(), {index}
        else:
            childprobs = [self.get_weight(c) for c in node.children]
            if ntype == 'conj':
                p = self.semiring.one()
                all_used = set()
                for cp, cu in childprobs:
                    all_used |= cu
                    p = self.semiring.times(p, cp)
                return p, all_used
            elif ntype == 'disj':
                p = self.semiring.zero()
                all_used = set()
                for cp, cu in childprobs:
                    all_used |= cu

                for cp, cu in childprobs:
                    not_used = all_used - cu
                    for nu in not_used:
                        nu_p, u = self.get_weight(nu)
                        nu_n, u = self.get_weight(-nu)
                        cp = self.semiring.times(cp, self.semiring.plus(nu_p, nu_n))
                    p = self.semiring.plus(p, cp)
                return p, all_used
            else:
                raise TypeError("Unexpected node type: '%s'." % ntype)



"""
problog.parser - Parser for Prolog programs
-------------------------------------------

Efficient low-level parser for Prolog programs.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from __future__ import print_function

from .errors import ParseError as CoreParseError

LINE_COMMENT = '%'
BLOCK_COMMENT_START = '/*'
BLOCK_COMMENT_END = '*/'
NEWLINE = '\n'

WHITESPACE = frozenset('\n\t ')


class ParseError(CoreParseError):
    def __init__(self, string, message, location):
        line, col, text = self._convert_pos(string, location)
        CoreParseError.__init__(self, message, location=(None, line, col), line=text)

    def _convert_pos(self, string, location):
        """Find line number, column number and text of offending line."""
        lineno = 1
        col = 0
        end = 0
        stop = False
        for i, x in enumerate(string):
            if x == '\n':
                if stop:
                    break
                lineno += 1
                col = 0
            if i == location:
                stop = True
            if not stop:
                col += 1
        return lineno, col, string[location - col:i + 1]


class UnexpectedCharacter(ParseError):
    def __init__(self, string, position):
        char = string[position]
        ParseError.__init__(self, string, "Unexpected character '%s'" % char, position)


class UnmatchedCharacter(ParseError):
    def __init__(self, string, position, length=1):
        char = string[position:position + length]
        ParseError.__init__(self, string, "Unmatched character '%s'" % char, position)


class Token(object):
    def __init__(self, string, pos, types=None, end=None, atom=True, functor=False, binop=None,
                 unop=None, special=None, atom_action=None):
        #        if end == None : end = pos+len(string)
        self.string = string
        self.location = pos
        self.atom = atom
        self.binop = binop
        self.unop = unop
        self.special = special
        self.arglist = False
        if atom:
            self.functor = functor
        else:
            self.functor = False
        self.is_comma_list = False
        self.atom_action = atom_action

    def is_special(self, special):
        return self.special == special

    def is_atom(self):
        return self.atom

    @property
    def priority(self):
        prior = 0
        if self.binop:
            prior = self.binop[0]
        elif self.unop:
            prior = self.unop[0]
        return prior

    def count_options(self):
        o = 0
        if self.atom:
            o += 1
        if self.binop:
            o += 1
        if self.unop:
            o += 1
        if self.functor:
            o += 1
        return o

    def list_options(self):  # pragma: no cover
        o = ''
        if self.atom:
            o += 'a'
        if self.binop:
            o += 'b'
        if self.unop:
            o += 'u'
        if self.functor:
            o += 'f'
        if self.arglist:
            o += 'l'
        return o

    def __repr__(self):  # pragma: no cover
        return "'%s' {%s}" % (self.string, self.list_options())


from collections import namedtuple

SPECIAL_PAREN_OPEN = 0
SPECIAL_PAREN_CLOSE = 1
SPECIAL_END = 2
SPECIAL_COMMA = 3
SPECIAL_BRACK_OPEN = 4
SPECIAL_BRACK_CLOSE = 5
SPECIAL_VARIABLE = 6
SPECIAL_FLOAT = 7
SPECIAL_INTEGER = 8
SPECIAL_PIPE = 9
SPECIAL_STRING = 10
SPECIAL_ARGLIST = 11

import re

RE_FLOAT = re.compile(r'[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?')


def skip_to(s, pos, char):
    end = s.find(char, pos)
    if end == -1:
        return len(s)
    else:
        return end + 1


def skip_comment_c(s, pos):
    end = s.find(BLOCK_COMMENT_END, pos)
    if end == -1:
        raise UnmatchedCharacter(s, pos, 2)
    return end + 2


def skip_comment_line(s, pos):
    return skip_to(s, pos, NEWLINE)


def is_lower(c):
    return 'a' <= c <= 'z'


def is_upper(c):
    return 'A' <= c <= 'Z'


def is_digit(c):
    return '0' <= c <= '9'


def is_whitespace(c):
    return c <= ' '


def is_comment_start(c):
    return c == '%' or c == '/'


class PrologParser(object):
    def __init__(self, factory):
        self.factory = factory
        self.prepare()

    def _skip(self, s, pos):
        return None, pos + 1

    def _next_paren_open(self, s, pos):
        try:
            return s[pos + 1] == '('
        except IndexError:
            return False

    def _token_notsupported(self, s, pos):
        raise UnexpectedCharacter(s, pos)

    def _token_dquot(self, s, pos):
        end = s.find('"', pos + 1)
        while end != -1 and s[end - 1] == '\\':
            end = s.find('"', end + 1)
        if end == -1:
            raise UnmatchedCharacter(s, pos)
        else:
            return Token(s[pos:end + 1], pos, special=SPECIAL_STRING), end + 1

    def _token_pound(self, s, pos):
        return Token(s[pos], pos, binop=(500, 'yfx', self.factory.build_binop),
                     functor=self._next_paren_open(s, pos)), pos + 1

    def _token_percent(self, s, pos):
        return None, skip_comment_line(s, pos)

    def _token_squot(self, s, pos):
        end = s.find("'", pos + 1)
        while end != -1 and s[end - 1] == '\\':
            end = s.find("'", end + 1)
        if end == -1:
            raise UnmatchedCharacter(s, pos)
        else:
            return Token(s[pos:end + 1], pos, functor=self._next_paren_open(s, end)), end + 1

    def _token_paren_open(self, s, pos):
        return Token(s[pos], pos, atom=False, special=SPECIAL_PAREN_OPEN), pos + 1

    def _token_paren_close(self, s, pos):
        return Token(s[pos], pos, atom=False, special=SPECIAL_PAREN_CLOSE), pos + 1

    def _token_asterisk(self, s, pos):
        if s[pos:pos + 2] == '**':
            return Token('**', pos, binop=(200, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        elif s[pos:pos + 3] == '*->':
            return Token('*->', pos, binop=(200, 'xfy', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 3
        else:
            return Token('*', pos, binop=(400, 'yfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 1

    def _token_plus(self, s, pos):
        return Token('+', pos, binop=(500, 'yfx', self.factory.build_binop),
                     unop=(200, 'fy', self.factory.build_unop),
                     functor=self._next_paren_open(s, pos)), pos + 1

    def _token_comma(self, s, pos):
        return Token(',', pos, binop=(1000, 'xfy', self.factory.build_conjunction), atom=False,
                     special=SPECIAL_COMMA), pos + 1

    def _token_min(self, s, pos):
        if s[pos:pos + 3] == '-->':
            return Token('-->', pos, binop=(1200, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 3
        elif s[pos:pos + 2] == '->':
            return Token('->', pos, binop=(1050, 'xfy', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        else:
            return Token('-', pos, binop=(500, 'yfx', self.factory.build_binop),
                         unop=(200, 'fy', self.factory.build_unop),
                         functor=self._next_paren_open(s, pos)), pos + 1

    def _token_dot(self, s, pos):
        if pos + 1 == len(s) or is_whitespace(s[pos + 1]) or is_comment_start(s[pos + 1]):
            return Token('.', pos, special=SPECIAL_END), pos + 1
        elif is_digit(s[pos + 1]):
            return self._token_number(s, pos)
        elif s[pos + 1] == '(':
            return Token('.', pos, functor=self._next_paren_open(s, pos)), pos + 1
        else:
            raise UnexpectedCharacter(s, pos)

    def _token_slash(self, s, pos):
        if s[pos:pos + 2] == '/\\':
            return Token('/\\', pos, binop=(500, 'yfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        elif s[pos:pos + 2] == '//':
            return Token('//', pos, binop=(400, 'yfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        elif s[pos:pos + 2] == '/*':
            return None, skip_comment_c(s, pos)
        else:
            return Token('/', pos, binop=(400, 'yfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 1

    def _token_colon(self, s, pos):
        if s[pos:pos + 2] == ':-':
            return Token(':-', pos, binop=(1200, 'xfx', self._build_clause),
                         unop=(1200, 'fx', self.factory.build_directive),
                         functor=self._next_paren_open(s, pos)), pos + 2
        elif s[pos:pos + 2] == '::':
            return Token('::', pos, binop=(1000, 'xfx', self.factory.build_probabilistic),
                         functor=self._next_paren_open(s, pos)), pos + 2
        else:
            return Token(':', pos, binop=(600, 'xfy', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 1

    def _token_semicolon(self, s, pos):
        return Token(';', pos, binop=(1100, 'xfy', self.factory.build_disjunction),
                     functor=self._next_paren_open(s, pos)), pos + 1

    def _token_less(self, s, pos):
        if s[pos:pos + 2] == '<-':
            return Token('<-', pos, binop=(1200, 'xfx', self._build_clause),
                         functor=self._next_paren_open(s, pos)), pos + 2
        elif s[pos:pos + 2] == '<<':
            return Token('<<', pos, binop=(400, 'yfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        else:
            return Token('<', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 1

    def _token_equal(self, s, pos):
        if s[pos:pos + 2] == '=<':
            return Token('=<', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        elif s[pos:pos + 3] == '=:=':
            return Token('=:=', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 3
        elif s[pos:pos + 3] == '=\=':
            return Token('=\=', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 3
        elif s[pos:pos + 3] == '=@=':
            return Token('=@=', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 3
        elif s[pos:pos + 3] == '=..':
            return Token('=..', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 3
        elif s[pos:pos + 2] == '==':
            return Token('==', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        else:
            return Token('=', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 1

    def _token_greater(self, s, pos):
        if s[pos:pos + 2] == '>>':
            return Token('>>', pos, binop=(400, 'yfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        elif s[pos:pos + 2] == '><':
            return Token('><', pos, binop=(500, 'yfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        elif s[pos:pos + 2] == '>=':
            return Token('>=', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        else:
            return Token('>', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 1

    def _token_question(self, s, pos):
        return Token('?', pos, atom=True), pos + 1
        # raise UnexpectedCharacter(s, pos)

    def _token_at(self, s, pos):
        if s[pos:pos + 2] == '@<':
            return Token('@<', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        elif s[pos:pos + 3] == '@=<':
            return Token('@=<', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 3
        elif s[pos:pos + 3] == '@>=':
            return Token('@>=', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 3
        elif s[pos:pos + 2] == '@>':
            return Token('@>', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        else:
            raise UnexpectedCharacter(s, pos)

    def _token_bracket_open(self, s, pos):
        return Token('[', pos, atom=False, special=SPECIAL_BRACK_OPEN), pos + 1

    def _token_backslash(self, s, pos):
        if s[pos:pos + 2] == '\\\\':
            return Token('\\\\', pos, unop=(200, 'fy', self.factory.build_unop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        elif s[pos:pos + 2] == '\\+':
            return Token('\\+', pos, unop=(900, 'fy', self.factory.build_not),
                         functor=self._next_paren_open(s, pos)), pos + 2
        elif s[pos:pos + 4] == '\\=@=':
            return Token('\\+', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 4
        elif s[pos:pos + 3] == '\\==':
            return Token('\\==', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 3
        elif s[pos:pos + 2] == '\\=':
            return Token('\\=', pos, binop=(700, 'xfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        elif s[pos:pos + 2] == '\\/':
            return Token('\\/', pos, binop=(500, 'yfx', self.factory.build_binop),
                         functor=self._next_paren_open(s, pos)), pos + 2
        else:
            return Token('\\', pos, unop=(200, 'fy', self.factory.build_unop),
                         functor=self._next_paren_open(s, pos)), pos + 1

    def _token_bracket_close(self, s, pos):
        return Token(']', pos, atom=False, special=SPECIAL_BRACK_CLOSE), pos + 1

    def _token_caret(self, s, pos):
        return Token('^', pos, binop=(400, 'xfy', self.factory.build_binop),
                     functor=self._next_paren_open(s, pos)), pos + 1

    def _token_underscore(self, s, pos):
        return self._token_upper(s, pos)  # Variable

    def _token_pipe(self, s, pos):
        return Token('|', pos, atom=False, binop=(1100, 'xfy', self.factory.build_binop),
                     special=SPECIAL_PIPE), pos + 1

    def _token_tilde(self, s, pos):
        return Token('~', pos, unop=(900, 'fx', self.factory.build_unop)), pos + 1

    def _token_lower(self, s, pos):
        end = pos + 1
        s_len = len(s)
        if end < s_len:
            c = s[end]
            while c == '_' or is_lower(c) or is_upper(c) or is_digit(c):
                end += 1
                if end >= s_len:
                    break
                c = s[end]
        token = s[pos:end]
        kwd = self.string_operators.get(token, {})
        return Token(token, pos, functor=self._next_paren_open(s, end - 1), **kwd), end

    def _token_upper(self, s, pos):
        end = pos + 1
        s_len = len(s)
        if end < s_len:
            c = s[end]
            while c == '_' or is_lower(c) or is_upper(c) or is_digit(c):
                end += 1
                if end >= s_len:
                    break
                c = s[end]
            return Token(s[pos:end], pos, special=SPECIAL_VARIABLE), end
        else:
            return Token(s[pos], pos, special=SPECIAL_VARIABLE), end

    def _token_number(self, s, pos):
        token = RE_FLOAT.match(s, pos).group(0)
        if token.find('.') >= 0 or token.find('e') >= 0 or token.find('E') >= 0:
            return Token(token, pos, special=SPECIAL_FLOAT), pos + len(token)
        else:
            return Token(token, pos, special=SPECIAL_INTEGER), pos + len(token)

    def _token_action(self, char):
        c = ord(char)
        if c < 33:
            return self._skip  # whitespace
        elif c < 48:
            return self._token_act1[c - 33]
        elif c < 58:
            return self._token_number
        elif c < 65:
            return self._token_act2[c - 58]
        elif c < 91:
            return self._token_upper
        elif c < 97:
            return self._token_act3[c - 91]
        elif c < 123:
            return self._token_lower
        elif c < 127:
            return self._token_act4[c - 123]
        else:
            return None

    def _build_clause(self, functor, operand1, operand2, location, **extra):
        heads = []
        current = operand1
        while current.functor == ';':
            heads.append(current.args[0])
            current = current.args[1]
        heads.append(current)
        return self.factory.build_clause(functor=functor, operand1=heads, operand2=operand2,
                                         location=location, **extra)

    def next_token(self, s, pos):
        action = self._token_action(s[pos])
        if action is None:
            raise UnexpectedCharacter(s, pos)
        result = action(s, pos)
        if result is None:  # pragma: no cover
            raise RuntimeError("Undefined action: '%s'" % action)
        else:
            return result

    def prepare(self):
        self._token_act1 = [
            self._token_notsupported,  # 33 !
            self._token_dquot,  # 34 "
            self._token_pound,  # 35 #
            self._token_notsupported,  # 36 $
            self._token_percent,  # 37 %
            self._token_notsupported,  # 38 &
            self._token_squot,  # 39 '
            self._token_paren_open,  # 40 (
            self._token_paren_close,  # 41 )
            self._token_asterisk,  # 42 *
            self._token_plus,  # 43 +
            self._token_comma,  # 44 ,
            self._token_min,  # 45 -
            self._token_dot,  # 46 .
            self._token_slash  # 47 /
        ]
        self._token_act2 = [
            self._token_colon,  # 58 :
            self._token_semicolon,  # 59 ;
            self._token_less,  # 60 <
            self._token_equal,  # 61 =
            self._token_greater,  # 62 >
            self._token_question,  # 63 ?
            self._token_at,  # 64 @
        ]
        self._token_act3 = [
            self._token_bracket_open,  # 91 [
            self._token_backslash,  # 92
            self._token_bracket_close,  # 93 ]
            self._token_caret,  # 94 ^
            self._token_underscore,  # 95 _
            self._token_notsupported,  # 96 `
        ]
        self._token_act4 = [
            self._token_notsupported,  # 123 {
            self._token_pipe,  # 124 |
            self._token_notsupported,  # 125 }
            self._token_tilde,  # 126 ~
        ]

        self.string_operators = {
            'is': {'binop': (700, 'xfx', self.factory.build_binop)},
            'not': {'unop': (900, 'fy', self.factory.build_not), 'atom': True},
            'xor': {'binop': (500, 'yfx', self.factory.build_binop)},
            'rdiv': {'binop': (400, 'yfx', self.factory.build_binop)},
            'mod': {'binop': (400, 'yfx', self.factory.build_binop)},
            'rem': {'binop': (400, 'yfx', self.factory.build_binop)},
            'div': {'binop': (400, 'yfx', self.factory.build_binop)}
        }

    def _tokenize(self, s):
        s_len = len(s)

        p = 0
        if s[:2] == '#!':
            p = skip_comment_line(s, p)
        while p < s_len:
            t, p = self.next_token(s, p)
            if t is not None:
                yield t

    def _extract_statements(self, string, s):
        statement = []

        for token in s:
            if token.is_special(SPECIAL_END):
                if not statement:
                    raise ParseError(string, 'Empty statement found', token.location)
                yield statement
                statement = []
            else:
                statement.append(token)
        if statement:
            raise ParseError(string, 'Incomplete statement', len(string))

    def _build_operator_free(self, string, tokens):
        if len(tokens) == 1:
            token = tokens[0]
            if isinstance(token, SubExpression):
                if isinstance(token.tokens, list):
                    curr = token.tokens[-1]
                    for t in reversed(token.tokens[:-1]):
                        curr = self.factory.build_conjunction(',', t, curr)
                    return curr
                else:
                    return token.tokens
            elif token.is_special(SPECIAL_VARIABLE):
                return self.factory.build_variable(token.string, location=token.location)
            elif token.is_special(SPECIAL_INTEGER):
                return self.factory.build_constant(int(token.string), location=token.location)
            elif token.is_special(SPECIAL_FLOAT):
                return self.factory.build_constant(float(token.string), location=token.location)
            elif token.is_special(SPECIAL_STRING):
                return self.factory.build_string(token.string[1:-1], location=token.location)
            else:
                return self.factory.build_function(token.string, (), location=token.location)
        elif len(tokens) == 2:
            args = [tok for tok in tokens[1].tokens]
            return self.factory.build_function(tokens[0].string, args, location=tokens[0].location)
        elif len(tokens) != 0:
            raise ParseError(string, 'Unexpected token', tokens[0].location)
        else:
            return None

    def fold(self, string, operators, lo, hi, pprior=None, porder=None, level=0):
        if lo >= hi:
            return self._build_operator_free(string, operators[lo:hi])
        else:
            max_op = None
            max_i = None
            for i in range(lo, hi):
                op_n = operators[i]
                op = None
                if op_n.binop:
                    op = op_n.binop
                elif op_n.unop:
                    op = op_n.unop
                if op is not None and (max_op is None or op[0] > max_op[0] or
                                       (op[0] == max_op[0] and max_op[1] == 'yfx')):
                    max_i = i
                    max_op = op
            if max_op is None:
                return self._build_operator_free(string, operators[lo:hi])
            else:
                if pprior == max_op[0] and porder == 'x':
                    raise ParseError(string, 'Operator priority clash', operators[max_i].location)
                else:
                    max_order = max_op[1]
                    if len(max_order) == 3:  # binop
                        lf = self.fold(string, operators, lo, max_i, max_op[0], max_order[0],
                                       level + 1)
                        rf = self.fold(string, operators, max_i + 1, hi, max_op[0], max_order[2],
                                       level + 1)
                        return max_op[2](functor=operators[max_i].string, operand1=lf, operand2=rf,
                                         location=operators[max_i].location,
                                         priority=max_op[0], opspec=max_op[1])
                    else:  # unop
                        if max_i != lo:
                            raise ParseError(string, 'Operator priority clash',
                                             operators[max_i].location)
                        lf = self.fold(string, operators, lo + 1, hi, max_op[0], max_order[1],
                                       level + 1)
                        return max_op[2](functor=operators[max_i].string, operand=lf,
                                         location=operators[max_i].location,
                                         priority=max_op[0], opspec=max_op[1])

    def label_tokens(self, string, tokens):
        l = len(tokens) - 1
        p = None
        for i, t in enumerate(tokens):
            if i == l:  # Last token can not be an operator or functor
                t.unop = None
                t.binop = None
                t.functor = False
            elif t.functor and tokens[i + 1].is_comma_list:
                t.atom = False
            elif t.unop and tokens[i + 1].priority > t.priority:
                t.unop = False

            if i == 0:
                t.binop = None  # First token can not be a binop
                t.arglist = False
            elif p.functor:
                t.atom = False
                t.arglist = t.is_comma_list
            elif p.arglist:
                t.unop = False
                t.atom = False
                t.functor = False
            elif p.atom:
                if not t.binop:
                    raise ParseError(string, 'Expected binary operator', t.location)
                t.unop = None
                t.atom = False
                t.functor = False
                t.arglist = False
            elif p.binop:
                t.binop = None
                t.arglist = False
            else:
                t.arglist = False

            if t.unop and t.functor:
                t.unop = None

            if t.unop and t.atom:
                n = tokens[i + 1]
                if not n.binop:
                    t.atom = False
            p = t
            if t.count_options() != 1:
                raise ParseError(string, 'Ambiguous token role', t.location)

        return tokens

    def _parse_statement(self, string, tokens):
        return self.collapse(string, tokens)

    def parseString(self, string):
        return self.factory.build_program(mapl(lambda x: self._parse_statement(string, x),
                                               self._extract_statements(string,
                                                                        self._tokenize(string))))

    def parseFile(self, filename):
        with open(filename) as f:
            return self.parseString(f.read())

    def collapse(self, string, tokens):
        """Combine tokens into subexpressions."""

        root_tokens = []
        expr_stack = []
        for token_i, token in enumerate(tokens):
            if token.is_special(SPECIAL_PAREN_OPEN):  # Open a parenthesis expression
                expr_stack.append(ParenExpression(string, token))
            elif token.is_special(SPECIAL_BRACK_OPEN):  # Open a list expression
                expr_stack.append(ListExpression(string, token))
            elif token.is_special(SPECIAL_PAREN_CLOSE) or token.is_special(SPECIAL_BRACK_CLOSE):
                try:
                    current_expr = expr_stack.pop(-1)  # Close a parenthesis expression
                    if not current_expr.accepts(token):
                        raise UnexpectedCharacter(string, token.location)
                    else:
                        current_expr.append(token)
                        current_expr.parse(self)
                        if not expr_stack:
                            root_tokens.append(current_expr)
                        else:
                            expr_stack[-1].append(current_expr)
                except IndexError:
                    raise UnmatchedCharacter(string, token.location)
            elif expr_stack:
                expr_stack[-1].append(token)
            else:
                root_tokens.append(token)
        if expr_stack:
            raise UnmatchedCharacter(string, expr_stack[-1].start.location)

        toks = self.label_tokens(string, root_tokens)
        return self.fold(string, toks, 0, len(toks))


def mapl(f, l):
    return list(map(f, l))


def filterl(f, l):
    return list(filter(f, l))


class SubExpression(object):
    def __init__(self, string, start):
        self.string = string
        self.tokens = []
        self.start = start
        self.end = None

        self.binop = None
        self.unop = False
        self.functor = False
        self.atom = True
        self._arglist = True
        self.max_operators = []

        self.priority = 0

    @property
    def arglist(self):
        return self._arglist and self.is_comma_list

    @arglist.setter
    def arglist(self, value):
        self._arglist = value

    @property
    def location(self):
        return self.start.location

    def parse(self, parser):
        self.label_tokens(parser)
        if self.arglist:
            current = []
            tokens = []
            for token in self.tokens:
                if token.is_special(SPECIAL_COMMA):
                    tokens.append(parser.fold(self.string, current, 0, len(current)))
                    current = []
                else:
                    current.append(token)
            new_token = parser.fold(self.string, current, 0, len(current))
            tokens.append(new_token)
            self.tokens = tokens
        else:
            self.tokens = parser.fold(self.string, self.tokens, 0, len(self.tokens))

    def label_tokens(self, parser):
        parser.label_tokens(self.string, self.tokens)

    def count_options(self):
        return 1

    def is_special(self, special):
        return False

    def append(self, token):
        if token.is_special(self.close_char):
            self.end = token
        elif token.binop:
            if not self.max_operators or token.binop[0] > self.max_operators[0].binop[0]:
                self.max_operators = [token]
            elif self.max_operators and token.binop[0] == self.max_operators[0].binop[0]:
                self.max_operators.append(token)
            self.tokens.append(token)
        else:
            self.tokens.append(token)

    def list_options(self):  # pragma: no cover
        o = ''
        if self.atom:
            o += 'a'
        if self.binop:
            o += 'b'
        if self.unop:
            o += 'u'
        if self.functor:
            o += 'f'
        if self.arglist:
            o += 'l'
        return o


class ListExpression(SubExpression):
    def __init__(self, string, start):
        SubExpression.__init__(self, string, start)
        self.close_char = SPECIAL_BRACK_CLOSE
        self.is_comma_list = False

    def accepts(self, token):
        return not token.is_special(SPECIAL_PAREN_CLOSE)

    def __repr__(self):
        return 'LE %s {%s}' % (self.tokens, self.list_options())

    def parse(self, parser):
        self.label_tokens(parser)
        prefix = []
        tail = None
        current = []
        for token_i, token in enumerate(self.tokens):
            if token.is_special(SPECIAL_PIPE):
                prefix.append(parser.fold(self.string, current, 0, len(current)))
                current = []
                tail = parser.fold(self.string, self.tokens[token_i + 1:], 0,
                                   len(self.tokens[token_i + 1:]))
                break
            elif token.is_special(SPECIAL_COMMA):
                prefix.append(parser.fold(self.string, current, 0, len(current)))
                current = []
            else:
                current.append(token)
        if current:
            prefix.append(parser.fold(self.string, current, 0, len(current)))
        self.tokens = parser.factory.build_list(prefix, tail)
        # else :
        #     self.tokens = parser.fold(self.string, self.tokens, 0, len(self.tokens) )


class ParenExpression(SubExpression):
    def __init__(self, string, tokens):
        SubExpression.__init__(self, string, tokens)
        self.close_char = SPECIAL_PAREN_CLOSE

    @property
    def is_comma_list(self):
        return not self.max_operators or self.max_operators[0].string == ',' \
            or self.max_operators[0].priority < 1000

    def accepts(self, token):
        return not token.is_special(SPECIAL_BRACK_CLOSE)

    def __repr__(self):
        return 'PE %s {%s}' % (self.tokens, self.list_options())


class Factory(object):
    """Factory object for creating suitable objects from the parse tree."""

    def build_program(self, clauses):
        return '\n'.join(map(str, clauses))

    def build_function(self, functor, arguments, location=None):
        return '%s(%s)' % (functor, ', '.join(map(str, arguments)))

    def build_variable(self, name, location=None):
        return str(name)

    def build_constant(self, value, location=None):
        return str(value)

    def build_binop(self, functor, operand1, operand2, function=None, location=None, **extra):
        return self.build_function("'" + functor + "'", (operand1, operand2), location=location)

    def build_unop(self, functor, operand, location=None, **extra):
        return self.build_function("'" + functor + "'", (operand,), location=location)

    def build_list(self, values, tail=None, location=None, **extra):
        if tail is None:
            return '[%s]' % (', '.join(map(str, values)))
        else:
            return '[%s | %s]' % (', '.join(map(str, values)), tail)

    def build_string(self, value, location=None):
        return self.build_constant('"' + value + '"', location=location)

    def build_cut(self, location=None):
        raise NotImplementedError('Not supported!')

    build_clause = build_binop
    build_probabilistic = build_binop
    build_disjunction = build_binop
    build_conjunction = build_binop
    build_compare_arithmetic = build_binop
    build_compare_struct = build_binop
    build_compare_eq = build_binop
    build_mathop2 = build_binop
    build_ifthen = build_binop
    build_list_op = build_binop

    build_not = build_unop
    build_mathop1 = build_unop
    build_directive = build_unop


def main(filenames):
    for filename in filenames:
        print(filename)
        print('------------------------------------')
        from problog.program import ExtendedPrologFactory

        try:
            parsed = PrologParser(ExtendedPrologFactory()).parseFile(filename)
            for s in parsed:
                print(s)
        except ParseError as e:
            print('ParseError:', e)
        print('====================================')


DefaultPrologParser = PrologParser

# from .parser_pyparsing import PrologParser as DefaultPrologParser
